/*
 * $Id: g_dihfix ,v  2007/07/7  A. Rzepiela  a.rzepiela@rug.nl  Exp $
 * 
 * 
 *  Reverse transformation tool based  on G   R   O   M   A   C   S
 * 
 *                       VERSION 3.3.1
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * If you want to redistribute modifications, please consider that
 * scientific software is very special. Version control is crucial -
 * bugs must be traceable. We will be happy to consider code for
 * inclusion in the official distribution, but derived work must not
 * be called official GROMACS. Details are found in the README & COPYING
 * files - if they are missing, get the official version at www.gromacs.org.
 * 
 * 
 */

#include <sys/types.h>
#include <math.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <time.h>
#include "sysstuff.h"
#include "smalloc.h"
#include "macros.h"
#include "string2.h"
#include "../kernel/readir.h"
#include "../kernel/toputil.h"
#include "../kernel/topio.h"
#include "confio.h"
#include "../kernel/topcat.h"
#include "copyrite.h"
#include "../kernel/readir.h"
#include "symtab.h"
#include "names.h"
#include "grompp.h"
#include "random.h"
#include "vec.h"
#include "futil.h"
#include "statutil.h"
#include "../kernel/splitter.h"
#include "sortwater.h"
#include "../kernel/convparm.h"
#include "fatal.h"
#include "index.h"
#include "gmxfio.h"
#include "trnio.h"
#include "tpxio.h"
#include "rmpbc.h"
#include "xtcio.h"
#include "../kernel/vsite_parm.h"
#include "txtdump.h"
#include "calcgrid.h"
#include "../kernel/add_par.h"
#include "enxio.h"
//#include <gmx_dihfix.h>
#include "nrama.h"
#include "bondf.h"
#include "physics.h"
#include "nrama.h"
#include <stdio.h>

static int find_min(real phi,int ntab,real phitab[])
{
  int  i,imin;
  real mind,mm;
  real width;
 
  /* Set closest minimum to the first one */
  width=360.0/ntab;
  mind=fabs(phi-phitab[0]);
  if(mind >( (360*(ntab-1)/ntab) + width*0.5 ))
	  mind = fabs(mind-360);
  imin=0;
  for(i=1; (i<ntab); i++) {
    mm=fabs(phi-phitab[i]);
    if (mm < mind) {
      imin=i;
      mind=mm;
    }
  }
  if (mind < width*0.5 ){
	  return imin;}
    
  else{
	  return -1;
    
  }
}

static int vphi(real phi,int mult) 
{
  static real m2[] = { 0, 180 };
  static real m3[] = { 0, 120, 240 };
  static real m4[] = { 0, 90, 180, 270 };
  static real m6[] = { 0, 60, 120, 180, 240, 300 };

  real phiref;
  int  vpp=0;
  
  phiref=RAD2DEG*phi;
  while (phiref < 0)
    phiref+=360;
  while (phiref > 360)
    phiref-=360;
  
  switch(mult) {
  case 2:
    vpp=find_min(phiref,2,m2);
    vpp=m2[vpp];
    break;
  case 3:
    vpp=find_min(phiref,3,m3);
    vpp=m3[vpp];
    break;
  case 4:
    vpp=find_min(phiref,4,m4);
    vpp=m4[vpp];
    break;
  case 6:
    vpp=find_min(phiref,6,m6);
    vpp=m6[vpp];
    break;
  default:
    gmx_fatal(FARGS,"No such multiplicity %d",mult);
  }

  if (vpp == -1)
    gmx_fatal(FARGS,"Couldn't calculate dihedral minimum");
  else
    return vpp;
}

static int *top_maker ( char *topfile,t_gromppopts *opts,t_inputrec *ir,
		bool bVerbose,t_atomtype *atype,t_topology *sys,
		t_molinfo *msys,t_params plist[],int *comb,
		real *reppow,bool bEnsemble,int *nerror )
{
	t_molinfo   *molinfo=NULL;
	t_simsystem *Sims=NULL;
	int         i,nrmols,Nsim,nmismatch;
	int         ntab,*tab;
	char        buf[STRLEN];

	init_molinfo ( msys );
	msys->name=do_top ( bVerbose,topfile,NULL,opts,& ( sys->symtab ),
			plist,comb,reppow,atype,&nrmols,&molinfo,ir,&Nsim,&Sims );
	ntab = 0;
	tab  = NULL;
	topcat ( msys,nrmols,molinfo,ntab,tab,Nsim,Sims,bEnsemble );
	mi2top ( sys,msys );
	
	return 0;
}
int gmx_dihfix ( int argc,char *argv[] )
{

	static char *desc[] =
	{
	 "Definition of dihedral angle restraints for ceretain dihedrals within \nthe peptide. For every dihedral with force constant larger than specified \nwith option fc, the dihedral restraint, that restraints it to the minimum \nthat is closest to the dihedral angle value from provided reference structure, will be constructed. Topology of peptide/protein without different molecules plus corresponding gro file should be provided "	 
	};

	static int   status, flags = TRX_READ_X, comb_fg, nerror_fg =0, minimum, i, printed=0;
	t_topology   *top_fg;
	char         title_fg[STRLEN],  fn_fg[STRLEN], *mdparin_fg, trajectory[STRLEN];
	t_trxframe   fr;
	rvec         *x_fg;
	static bool  bVerbose=TRUE;
	t_molinfo    msys_fg;
	real         repow_fg, fc=10, width=60;
	t_atomtype   atype_fg;
	t_params     *plist_fg;
	t_inputrec   *ir_fg;
	t_gromppopts *opts_fg;
        t_xrama      *xr;
	FILE *infile, *out;
	char line_buffer[BUFSIZ];
	t_pargs pa[] =
	{
		{ "-fc", FALSE, etREAL, {&fc},
			"Every dihedral with force constant larger than fc is listed in [dihedral_restraints] section, dafault 10"}, 
         	{ "-w", FALSE, etREAL, {&width},
			"Width*multiplicity for dihedral restraints, default 60 "} 
	};
	t_filenm fnm[] =
	{
		{ efTOP,  "-p",  NULL, ffREAD  },  /* fg topology file */
		{ efTRX, "-c", NULL, ffREAD },       /* fg structure file */
		{ efTOP, "-o","out", ffWRITE },      /* output top file */

	};
	CopyRight(stderr,argv[0]);
	#define NFILE asize(fnm)
	parse_common_args ( &argc,argv,0,NFILE,fnm,asize ( pa ),pa,asize ( desc ),desc,0,NULL );
	snew ( ir_fg,1 );
	snew ( opts_fg,1 );
	init_ir ( ir_fg,opts_fg );
	opts_fg->cpp =strdup ( "cpp" );
	snew(xr,1);
	/* prep top */
	snew ( top_fg,1 );
	init_top ( top_fg );
	init_molinfo ( &msys_fg );
	strcpy ( fn_fg,opt2fn ( "-p",NFILE,fnm ) );
	/* plist part */
	snew ( plist_fg,F_NRE );
	init_plist ( plist_fg );
        
	infile = fopen(opt2fn ( "-p",NFILE,fnm ), "r");
	char *bf;
        if (fexist(opt2fn ( "-o",NFILE,fnm ))) {
		bf=(char *)backup_fn(opt2fn ( "-o",NFILE,fnm ));
	        rename(opt2fn ( "-o",NFILE,fnm ),bf);}
	out = fopen(opt2fn ( "-o",NFILE,fnm ), "w");	
	if (!infile) {
		            printf("Couldn't open file %s for reading.\n", opt2fn ( "-p",NFILE,fnm ));
			            return 0;
		     }
	/* reading top */
	top_maker ( fn_fg,opts_fg,ir_fg,bVerbose,&atype_fg,top_fg,&msys_fg, plist_fg,
			&comb_fg,&repow_fg, ( opts_fg->eDisre==edrEnsemble ),&nerror_fg );
	convert_params(atype_fg.nr, plist_fg, msys_fg.plist, comb_fg, repow_fg, &top_fg->idef);
	/* check part */
	printf ( "\nNumber of atoms in protein %d \n", top_fg->atoms.nr );
	init_rama_rev(ftp2fn ( efTRX,NFILE,fnm ),top_fg,xr,3);
	while (fgets(line_buffer, sizeof(line_buffer), infile)) {
          if(strncmp(line_buffer,"[ dihedrals ]", 4)==0 && printed==0){
		fprintf(out, "[dihedral_restraints]\n");
		for(i =0; i<(xr->ndih); i++)
		{
			if(xr->dih[i].cpA > fc){
			minimum = vphi(xr->dih[i].ang,xr->dih[i].mult);
			fprintf(out , "%d %d %d %d 1 1 %d %f 1 2 \n", (xr->dih[i].ai[0]+1), (xr->dih[i].ai[1]+1),(xr->dih[i].ai[2]+1),(xr->dih[i].ai[3]+1),minimum, (width/xr->dih[i].mult)); 
		}
		}
		fprintf(out ,"\n");
	        printed =1;
	}
	fprintf(out,"%s",  line_buffer);
	}
	printf("\nDone!\n");
	fclose(infile);
	fclose(out);
	return 0;
	thanx ( stderr );

}
