/*
 * $Id: g_fg2cg  2007/07/7 A Rzepiela a.rzepiela@rug.nl  Exp $
 * 
 *  Reverse transformation tool based  on G   R   O   M   A   C   S
 * 
 *                       VERSION 3.3.1
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * If you want to redistribute modifications, please consider that
 * scientific software is very special. Version control is crucial -
 * bugs must be traceable. We will be happy to consider code for
 * inclusion in the official distribution, but derived work must not
 * be called official GROMACS. Details are found in the README & COPYING
 * files - if they are missing, get the official version at www.gromacs.org.
 * 
 * 
 */
#include <sys/types.h>
#include <math.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <time.h>
#include "sysstuff.h"
#include "smalloc.h"
#include "macros.h"
#include "string2.h"
#include "../kernel/readir.h"
#include "../kernel/toputil.h"
#include "../kernel/topio.h"
#include "confio.h"
#include "../kernel/topcat.h"
#include "copyrite.h"
#include "../kernel/readir.h"
#include "symtab.h"
#include "names.h"
#include "grompp.h"
#include "random.h"
#include "vec.h"
#include "futil.h"
#include "statutil.h"
#include "../kernel/splitter.h"
#include "sortwater.h"
#include "../kernel/convparm.h"
#include "fatal.h"
#include "index.h"
#include "gmxfio.h"
#include "trnio.h"
#include "tpxio.h"
#include "rmpbc.h"
#include "xtcio.h"
#include "../kernel/vsite_parm.h"
#include "txtdump.h"
#include "calcgrid.h"
#include "../kernel/add_par.h"
#include "enxio.h"
//#include "gmx_fg2cg.h"
#include "nrama.h"
#include "bondf.h"
#include "physics.h"

void  compute_cg_center ( t_topology *top_fg, rvec x_fg[],  rvec x_cg[], int nrAtoms_cg, matrix box )
{

	int i, j, jstart, kk, dd;
	int k;
	real m, m1, dist_to_center;
	real *mAux, Aux2;
	rvec Aux, *temp;
	rm_pbc(&(top_fg->idef),top_fg->atoms.nr, box ,x_fg,x_fg);
	for ( i = 0; i <  nrAtoms_cg ; i++ )
	{     	
		clear_rvec(x_cg[i]);
		m =0;
		for ( j = top_fg->atoms.mapp.index[i]; j < top_fg->atoms.mapp.index[i+1]; j ++ )
		{
			k= top_fg->atoms.mapp.a[j];
			clear_rvec ( Aux );
			svmul ( top_fg -> atoms.atom[k].m, x_fg[k], Aux );
			rvec_add ( Aux, x_cg[i], x_cg[i] );
			m  = m + top_fg -> atoms.atom[k].m;
		}
		m1 = 1/m;
		svmul ( m1, x_cg[i] , x_cg[i] );
		
	}
}


static int *top_maker ( char *topfile,t_gromppopts *opts,t_inputrec *ir,
		bool bVerbose,t_atomtype *atype,t_topology *sys,
		t_molinfo *msys,t_params plist[],int *comb,
		real *reppow,bool bEnsemble,int *nerror )
{
	t_molinfo   *molinfo=NULL;
	t_simsystem *Sims=NULL;
	int         i,nrmols,Nsim,nmismatch;
	int         ntab,*tab;
	char        buf[STRLEN];

	init_molinfo ( msys );
	msys->name=do_top ( bVerbose,topfile,NULL,opts,& ( sys->symtab ),
			plist,comb,reppow,atype,&nrmols,&molinfo,ir,&Nsim,&Sims );
	ntab = 0;
	tab  = NULL;
	topcat ( msys,nrmols,molinfo,ntab,tab,Nsim,Sims,bEnsemble );
	mi2top ( sys,msys );
	
	return 0;
}
void put_random ( rvec *x_fg, rvec x_cg, t_topology *top_fg, int  k, real radius, matrix box )
{
	int m = 0;
	long minus;
	real distance;
	rvec x, dx, cg_temp;
	t_pbc pbc;
	set_pbc ( &pbc,box );
	clear_rvec ( cg_temp );
	clear_rvec ( x );
	for ( m=0; ( m<DIM ); m++ )
	{
		distance = 0;
		int a = 0;
		while ( distance > radius || distance == 0 )
		{
			x[m]  = ( float ) drand48();
			minus = mrand48();
			if ( minus < 0 )
				x[m]= x[m]* ( -1 );
			x[m] +=  x_cg[m];
			if ( m == 0 )
			{
				cg_temp[m] = x_cg[m];
				pbc_dx ( &pbc , x , cg_temp, dx );
			}
			if ( m == 1 )
			{
				cg_temp[m-1] = x_cg[m-1], cg_temp[m] = x_cg[m];
				pbc_dx ( &pbc , x , cg_temp, dx );
			}
			if ( m == 2 )
			{
				pbc_dx ( &pbc , x , x_cg, dx );
			}
			distance = fabsf ( norm ( dx ) );
			a++;
		}
		x_fg[k][m] = x[m];
	}
}
void put_fg_atoms ( rvec *x_fg, rvec x_cg[], t_topology *top_fg, t_topology *top_cg, matrix box, real radius )
{
	int h, i, j, j1, k, k1,bone=0, m , res, max_backbone=0, backbone=0, *back , atoms[2], fg_list[5];
	long minus;
	rvec x, d;
	real distance = 0, amino_mass[12], amino_mass_cg, alpha_final_Y= 0 , alpha_final_Z = 0 ;
	static	int	first_time = TRUE;
	rvec W[3], trans, w_cg, wat_fg[12], Aux, AMINO[9], amino_cg;
	char name[2];
	/* Water */
	W[0][XX]= 0.569,W[0][YY]= 1.275, W[0][ZZ]= 1.165, W[1][XX]= 0.476, W[1][YY]= 1.268;
	W[1][ZZ]= 1.128, W[2][XX]= 0.580, W[2][YY]= 1.364, W[2][ZZ]= 1.209;

	for ( i =0; i<4; i++ )
	{
		trans[XX]=0.0, trans[YY]=0.0, trans[ZZ]=0.0;
		if ( i==1 )
			trans[XX]=0.2;
		if ( i==2 )
			trans[YY]=0.2;
		if ( i==3 )
		{
			trans[XX]=0.2;
			trans[YY]=0.2;
		}
		for ( j= 0; j< 3 ; j++ )
		{
			rvec_add ( trans, W[j], wat_fg[i*3+j] );
		}
	}
	clear_rvec ( w_cg );
	clear_rvec ( Aux );
	for ( h=0; h< 12; h++ )
	{
		if ( ( h%3 ) ==0 )
			svmul ( 16, wat_fg[h], Aux );
		else
			svmul ( 1, wat_fg[h], Aux );
		rvec_add ( w_cg, Aux, w_cg );
	}
	svmul ( 0.0138889, w_cg, w_cg );
	/* End water */


	/* Main part */
	if ( first_time )
	{
		first_time = FALSE;
		srand48 ( ( unsigned int ) time ( NULL ) );
		printf ( "%d \n" , ( unsigned int ) time ( NULL ) );
	}
	printf("Name of CG water in topology should be W\n"); 
	for ( i =0; i < top_cg->atoms.nr ; i++ )
	{
		/* Water */
		if ( strcmp ( * ( top_cg->atoms.atomname[i] ), "W" ) == 0 )
		{
			clear_rvec ( d );
			k1=0;
			d[XX] = x_cg[i][XX] - w_cg[XX];
			d[YY] = x_cg[i][YY] - w_cg[YY];
			d[ZZ] = x_cg[i][ZZ] - w_cg[ZZ];
			for ( j1 = top_fg->atoms.mapp.index[i]; j1 < top_fg->atoms.mapp.index[i+1]; j1 ++ )
			{
				k= top_fg->atoms.mapp.a[j1];
				rvec_add ( wat_fg[k1], d, x_fg[k] );
				k1++;
			}
		}
		else
		{
			for ( j1 = top_fg->atoms.mapp.index[i]; j1 < top_fg->atoms.mapp.index[i+1]; j1 ++ )
			{
				k= top_fg->atoms.mapp.a[j1];
				put_random ( x_fg, x_cg[i],top_fg, k, radius, box );
			}
		}
	}
}

int gmx_fg2cg ( int argc,char *argv[] )
{

	static char *desc[] =
	{
		"The tool translates FG structure to CG structure or the other way round \n -mapping in fg topology file must be declared \n -watermolecules must be coarse grain ready if you want to make a CG water \n -wat option rewrites FG-CG 12 atom water molecule to normal spc water. You have to specify FG topology and input gro file "
	};

	static int n=1, wat =0, status, flags = TRX_READ_X, comb_fg,comb_cg, nerror_fg =0, nerror_cg =0;
	t_topology *top_fg, *top_cg;
	char       title_fg[STRLEN], title_cg[STRLEN],  fn_fg[STRLEN], *mdparin_fg, fn_cg[STRLEN], *mdparin_cg, trajectory[STRLEN];
	t_trxframe fr;
	rvec       *x_fg, *x_cg;
	static bool bVerbose=TRUE;
	t_molinfo msys_fg, msys_cg;
	real repow_fg, repow_cg, radius=0.3;
	t_atomtype   atype_fg, atype_cg;
	t_params *plist_fg, *plist_cg;

	t_inputrec *ir_fg, *ir_cg;
	t_gromppopts *opts_fg, *opts_cg;


	t_pargs pa[] =
	{
		{ "-n", FALSE, etINT, {&n},
			"1: fg2cg transformation, 0: cg2fg transformation"
		},
		{ "-wat", FALSE, etINT, {&wat},
			"1: rewrites FG_CG water to true fg water; "
		},
		{ "-rad", FALSE, etREAL, {&radius},
			"A radius for random atom insertion; "
		}

	};


	t_filenm fnm[] =
	{
		{ efTOP,  "-pfg",  NULL, ffREAD  },  /* fg topology file */
		{ efTOP,  "-pcg",  NULL, ffOPTRD },  /* cg topology file */
		{ efTRX, "-c", NULL, ffREAD },       /* fg structure file */
		{ efSTO, "-o","out",  ffWRITE }, /* output gro file */

	};
	CopyRight(stderr,argv[0]);
#define NFILE asize(fnm)
	parse_common_args ( &argc,argv,0,NFILE,fnm,asize ( pa ),pa,asize ( desc ),desc,0,NULL );
	snew ( ir_fg,1 ), snew ( ir_cg,1 );
	snew ( opts_fg,1 ), snew ( opts_cg,1 );
	init_ir ( ir_fg,opts_fg ), init_ir ( ir_cg,opts_cg );
	opts_fg->cpp =strdup ( "cpp" );
	opts_cg->cpp =strdup ( "cpp" );

	/* prep top */
	snew ( top_fg,1 ), snew ( top_cg,1 );
	init_top ( top_fg ), init_top ( top_cg );
	init_molinfo ( &msys_fg ), init_molinfo ( &msys_cg );
	strcpy ( fn_fg,opt2fn ( "-pfg",NFILE,fnm ) ), strcpy ( fn_cg,opt2fn ( "-pcg",NFILE,fnm ) ), strcpy ( trajectory,opt2fn ( "-o",NFILE,fnm ) );

	/* plist part */
	snew ( plist_fg,F_NRE ), snew ( plist_cg,F_NRE );
	init_plist ( plist_fg ), init_plist ( plist_cg );

	/* reading top */
	top_maker ( fn_fg,opts_fg,ir_fg,bVerbose,&atype_fg,top_fg,&msys_fg, plist_fg,
			&comb_fg,&repow_fg, ( opts_fg->eDisre==edrEnsemble ),&nerror_fg );

	convert_params(atype_fg.nr, plist_fg, msys_fg.plist, comb_fg, repow_fg, &top_fg->idef);
	top_maker ( fn_cg,opts_cg,ir_cg,bVerbose,&atype_cg,top_cg,&msys_cg, plist_cg,
			&comb_cg,&repow_cg, ( opts_cg->eDisre==edrEnsemble ),&nerror_cg );

	/* check part */
	printf ( "\nNumber of fg atoms %d \n", top_fg->atoms.nr );
	printf ( "Number of cg atoms %d \n", top_cg->atoms.nr );

	/* reading .gro file */
	read_first_frame ( &status,ftp2fn ( efTRX,NFILE,fnm ),&fr,flags );
	if ( wat )
	{
		write_sto_conf ( ftp2fn ( efSTO,NFILE,fnm ),*top_fg->name,& ( top_fg->atoms ),fr.x,NULL,fr.box );
		printf ( "\n Water rewrited to normal representation  !\n" );
	}
	else
	{
		if ( n == 1 )
		{
			int frames = 0;
			int xfile = -1;
			int char_number = 0;
			char_number = strcspn( trajectory, ".");
			char xfilename2[100000];
			strncpy(xfilename2,trajectory , (char_number));
			strcat(xfilename2, ".xtc");	
			snew ( x_cg, top_cg->atoms.nr +1 );
			int trxout = -1;
			while ( read_next_frame ( status,&fr ) )
			{
				if ( frames > 0 )
				{
					if ( frames == 1 )
					{
						xfile = open_xtc ( xfilename2 , "w" );
					}
					write_xtc ( xfile, top_cg->atoms.nr, fr.step, fr.time, fr.box, x_cg, 1000 );
				}
				compute_cg_center ( top_fg, fr.x,  x_cg,  top_cg->atoms.nr, fr.box );
				frames++;
			}
			if ( frames > 0 )
				close_xtc(xfile);
			if ( frames == 0 )
			{
				compute_cg_center ( top_fg, fr.x,  x_cg,  top_cg->atoms.nr, fr.box );
				write_sto_conf ( ftp2fn ( efSTO,NFILE,fnm ),*top_cg->name,& ( top_cg->atoms ),x_cg,NULL,fr.box );
			}
			printf ( "\nCg structure computed !\n" );
		}
		else
		{
			snew ( x_fg, top_fg->atoms.nr +1 );
			put_fg_atoms ( x_fg, fr.x, top_fg, top_cg, fr.box, radius );
			printf ( "\nFg pre-structure computed !\n" );
			write_sto_conf ( ftp2fn ( efSTO,NFILE,fnm ),*top_fg->name,& ( top_fg->atoms ),x_fg,NULL,fr.box );
		}
	}
	return 0;
	thanx ( stderr );


}
